from typing import Any, Dict, List, Tuple

import structlog
from tqdm import tqdm

from rasa.llm_fine_tuning.conversations import Conversation
from rasa.llm_fine_tuning.paraphrasing.conversation_rephraser import (
    ConversationRephraser,
)
from rasa.llm_fine_tuning.paraphrasing.rephrase_validator import RephraseValidator
from rasa.llm_fine_tuning.paraphrasing.rephrased_user_message import (
    RephrasedUserMessage,
)
from rasa.llm_fine_tuning.storage import StorageContext
from rasa.shared.core.flows import FlowsList
from rasa.shared.exceptions import ProviderClientAPIException

PARAPHRASING_MODULE_STORAGE_LOCATION = "2_rephrasings"

structlogger = structlog.get_logger()


async def create_paraphrased_conversations(
    conversations: List[Conversation],
    rephrase_config: Dict[str, Any],
    num_rephrases: int,
    flows: FlowsList,
    llm_command_generator_config: Dict[str, Any],
    storage_context: StorageContext,
) -> Tuple[List[Conversation], Dict[str, Any]]:
    """Create paraphrased conversations.

    Rephrase all user messages of a conversation and divide them into passing
    and failing rephrasings.

    Args:
        conversations: The conversations.
        rephrase_config: The path to the rephrase configuration file.
        num_rephrases: The number of rephrases to produce per user message.
        flows: All flows.
        llm_command_generator_config: The configuration of the trained model.
        storage_context: The storage context.

    Returns:
        The conversations including rephrasings and the configuration used for
        rephrasing.
    """
    rephraser = ConversationRephraser(rephrase_config)
    validator = RephraseValidator(llm_command_generator_config, flows)

    if num_rephrases <= 0:
        structlogger.info(
            "paraphrasing_module.skip",
            num_rephrases=num_rephrases,
            message="Skipping paraphrasing module as user messages should not be "
            "rephrased.",
        )
        return conversations, rephraser.config

    rephrased_conversations: List[Conversation] = []
    for i in tqdm(range(len(conversations))):
        current_conversation = conversations[i]

        try:
            # rephrase all user messages even if rephrase=False is set
            # to not confuse the LLM and get valid output
            rephrasings = await rephraser.rephrase_conversation(
                conversations[i], num_rephrases
            )
            # filter out the rephrasings for user messages that have rephrase=False set
            rephrasings = _filter_rephrasings(rephrasings, conversations[i])
            # check if the rephrasings are still producing the same commands
            rephrasings = await validator.validate_rephrasings(
                rephrasings, current_conversation
            )
        except ProviderClientAPIException as e:
            structlogger.error(
                "paraphrasing_module.skip_conversation",
                conversation=current_conversation.name,
                exception=str(e),
            )
            continue

        for j, step in enumerate(
            current_conversation.iterate_over_annotated_user_steps(rephrase=True)
        ):
            step.passed_rephrasings = rephrasings[j].passed_rephrasings
            step.failed_rephrasings = rephrasings[j].failed_rephrasings

        rephrased_conversations.append(current_conversation)

        storage_context.write_conversations(
            conversations, PARAPHRASING_MODULE_STORAGE_LOCATION
        )

    return rephrased_conversations, rephraser.config


def _filter_rephrasings(
    rephrasings: List[RephrasedUserMessage], conversation: Conversation
) -> List[RephrasedUserMessage]:
    """Filter rephrasings.

    Return only those rephrasings for user messages that have rephrase=True.

    Args:
        rephrasings: All rephrased user messages of the conversation.
        conversation: The conversation.

    Returns:
        Rephrasings for those user messages that have rephrase=True.
    """
    filtered_rephrasings = []
    index = 0
    user_messages = conversation.get_user_messages_to_rephrase()

    for rephrasing in rephrasings:
        if index >= len(user_messages):
            break

        # the user messages and the rephrasings are in the same order
        # rephrasings might contain more user messages as the user messages that
        # should be rephrased
        if rephrasing.original_user_message == user_messages[index]:
            filtered_rephrasings.append(rephrasing)
            index += 1

    return filtered_rephrasings



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, List, Set, Text

from rasa.shared.constants import ACTION_ASK_PREFIX, UTTER_ASK_PREFIX
from rasa.shared.core.constants import (
    KEY_ASK_CONFIRM_DIGRESSIONS,
    KEY_BLOCK_DIGRESSIONS,
)
from rasa.shared.core.flows.flow_step import FlowStep
from rasa.shared.core.flows.utils import extract_digression_prop
from rasa.shared.core.slots import SlotRejection


@dataclass
class CollectInformationFlowStep(FlowStep):
    """A flow step for asking the user for information to fill a specific slot."""

    collect: str
    """The collect information of the flow step."""
    utter: str
    """The utterance that the assistant uses to ask for the slot."""
    collect_action: str
    """The action that the assistant uses to ask for the slot."""
    rejections: List[SlotRejection]
    """how the slot value is validated using predicate evaluation."""
    ask_before_filling: bool = False
    """Whether to always ask the question even if the slot is already filled."""
    reset_after_flow_ends: bool = True
    """Whether to reset the slot value at the end of the flow."""
    ask_confirm_digressions: List[str] = field(default_factory=list)
    """The flow id digressions for which the assistant should ask for confirmation."""
    block_digressions: List[str] = field(default_factory=list)
    """The flow id digressions that should be blocked during the flow step."""

    @classmethod
    def from_json(
        cls, flow_id: Text, data: Dict[str, Any]
    ) -> CollectInformationFlowStep:
        """Create a CollectInformationFlowStep object from serialized data.

        Args:
            flow_id: The id of the flow that contains the step.
            data: data for a CollectInformationFlowStep object in a serialized format

        Returns:
            A CollectInformationFlowStep object
        """
        base = super().from_json(flow_id, data)
        return CollectInformationFlowStep(
            collect=data["collect"],
            utter=data.get("utter", f"{UTTER_ASK_PREFIX}{data['collect']}"),
            # as of now it is not possible to define a different name for the
            # action, always use the default name 'action_ask_<slot_name>'
            collect_action=f"{ACTION_ASK_PREFIX}{data['collect']}",
            ask_before_filling=data.get("ask_before_filling", False),
            reset_after_flow_ends=data.get("reset_after_flow_ends", True),
            rejections=[
                SlotRejection.from_dict(rejection)
                for rejection in data.get("rejections", [])
            ],
            ask_confirm_digressions=extract_digression_prop(
                KEY_ASK_CONFIRM_DIGRESSIONS, data
            ),
            block_digressions=extract_digression_prop(KEY_BLOCK_DIGRESSIONS, data),
            **base.__dict__,
        )

    def as_json(self) -> Dict[str, Any]:
        """Serialize the CollectInformationFlowStep object.

        Returns:
            the CollectInformationFlowStep object as serialized data
        """
        data = super().as_json()
        data["collect"] = self.collect
        data["utter"] = self.utter
        data["ask_before_filling"] = self.ask_before_filling
        data["reset_after_flow_ends"] = self.reset_after_flow_ends
        data["rejections"] = [rejection.as_dict() for rejection in self.rejections]
        data["ask_confirm_digressions"] = self.ask_confirm_digressions
        data["block_digressions"] = (
            self.block_digressions if self.block_digressions else False
        )

        return data

    @property
    def default_id_postfix(self) -> str:
        """Returns the default id postfix of the flow step."""
        return f"collect_{self.collect}"

    @property
    def utterances(self) -> Set[str]:
        """Return all the utterances used in this step."""
        return {self.utter} | {r.utter for r in self.rejections}
